// Module for DirectInput
#include "app.h"

static IDirectInput *pDI=NULL; // DirectInput interface
static HWND hWnd=NULL; // The window handle to use for DirectInput

#define MAX_JOYSTICKS (64)
static IDirectInputDevice2 *JoyList[MAX_JOYSTICKS];
static IDirectInputDevice  *Keyboard=NULL;
// Number of joysticks connected to this machine
static int nJoystickCount=0;

static DIJOYSTATE *pCalcState=NULL;
#define DINPUT_BUTTON_COUNT (sizeof(pCalcState->rgbButtons)/sizeof(pCalcState->rgbButtons[0]))

// This function is called back with each joystick connected to the system
static BOOL CALLBACK DIEnumDevicesProc(LPCDIDEVICEINSTANCE pDDI,LPVOID pvRef)
{
  int nRet=0;
  IDirectInputDevice  *pDev =NULL;
  IDirectInputDevice2 *pDev2=NULL;
  (void)pvRef;

  if (nJoystickCount>=MAX_JOYSTICKS) return DIENUM_STOP; // Already have the maximum number of joysticks
  if (pDDI==NULL) return DIENUM_CONTINUE;

  // Create the DirectInput1 interface
  nRet=pDI->CreateDevice(pDDI->guidInstance,&pDev,NULL);
  if (nRet<0 || pDev ==NULL) return DIENUM_CONTINUE;

  // Get the DirectInputDevice2 interface so we can use the Poll method
  nRet=pDev->QueryInterface(IID_IDirectInputDevice2,(void **)&pDev2);
  RELEASE(pDev)
  if (nRet<0 || pDev2==NULL) return DIENUM_CONTINUE;

  // We now have a IDirectInputDevice2 interface to store
  JoyList[nJoystickCount++]=pDev2;

  return DIENUM_CONTINUE;
}

// Sets up one Joystick (for example the range of the joystick's axes)
static int SetUpJoystick(int i)
{
  int nRet=0; IDirectInputDevice2 *pJoy=NULL;
  DIPROPRANGE diprg;
  pJoy=JoyList[i];
  if (pJoy==NULL) return 1;

  nRet=pJoy->SetDataFormat(&c_dfDIJoystick);
  if (nRet<0) return 1;
  nRet=pJoy->SetCooperativeLevel(hWnd,DISCL_NONEXCLUSIVE|DISCL_BACKGROUND);

  // Make range is -0x10000 to 0x10000
  memset(&diprg,0,sizeof(diprg));
  diprg.diph.dwSize       = sizeof(diprg);
  diprg.diph.dwHeaderSize = sizeof(diprg.diph);
  diprg.diph.dwObj        = DIJOFS_X;
  diprg.diph.dwHow        = DIPH_BYOFFSET;
  diprg.lMin              = -0x10000;
  diprg.lMax              =  0x10000;
  nRet=pJoy->SetProperty(DIPROP_RANGE,&diprg.diph);
  diprg.diph.dwObj        = DIJOFS_Y;
  nRet=pJoy->SetProperty(DIPROP_RANGE,&diprg.diph);
  diprg.diph.dwObj        = DIJOFS_Z;
  nRet=pJoy->SetProperty(DIPROP_RANGE,&diprg.diph);
  nRet=pJoy->Acquire();

  return 0;
}

static int DirKeyboardInit()
{
  int nRet=0;
  // Get a DI interface to the keyboard
  Keyboard=NULL;
  nRet=pDI->CreateDevice(GUID_SysKeyboard,&Keyboard,NULL);
  if (nRet<0 || Keyboard==NULL) return 1;
  nRet=Keyboard->SetDataFormat(&c_dfDIKeyboard);
  nRet=Keyboard->SetCooperativeLevel(hWnd,DISCL_NONEXCLUSIVE|DISCL_BACKGROUND); 
  nRet=Keyboard->Acquire();
  return 0;
}

int DinpInit(HINSTANCE hInst,HWND hPassWnd)
{
  int nRet=0; int i=0;
  hWnd=hPassWnd;
  // Make sure all Interfaces are NULL
  nJoystickCount=0;
  memset(&JoyList,0,sizeof(JoyList));

  nRet=DirectInputCreate(hInst,DIRECTINPUT_VERSION,&pDI,NULL);
  if (nRet<0) return 1;

  // enumerate the joysticks connected to the system
  nRet=pDI->EnumDevices(DIDEVTYPE_JOYSTICK,DIEnumDevicesProc,NULL,DIEDFL_ATTACHEDONLY);
  

  // Set up the DI2 devices
  for (i=0;i<nJoystickCount;i++) SetUpJoystick(i);

  // Set up the keyboard
  DirKeyboardInit();

  return 0;
}

int DinpExit()
{
  int i=0;

  // Release all the Effect/DI2 interfaces
  for (i=0;i<MAX_JOYSTICKS;i++) { RELEASE(JoyList[i]) }
  nJoystickCount=0;

  // Release the keyboard interface
  RELEASE(Keyboard)

  // Release the DirectInput interface
  RELEASE(pDI)
  return 0;
}

// A flag to say whether a Joystick has been read this frame
static unsigned char bJoystickRead[MAX_JOYSTICKS];
static DIJOYSTATE JoyState[MAX_JOYSTICKS];
static int bKeyboardRead=0; // A flag to say if the Keyboard was read this frame
static unsigned char KeyboardState[256];
int DinpStart() // Call before checking for Input in a frame
{
  // No joysticks have been Read for this frame
  memset(&bJoystickRead,0,nJoystickCount*sizeof(bJoystickRead[0]));
  // Keyboard not Read this frame
  bKeyboardRead=0;
  return 0;
}

// Read one of the joysticks
static inline int ReadJoystick(int i)
{
  int j=0,nRet=0;
  IDirectInputDevice2 *pJoy=NULL;
  if (bJoystickRead[i]==1) return 0; // already read this frame - ready to go
  if (bJoystickRead[i]==2) return 1; // error reading joystick
  
  pJoy=JoyList[i];

  bJoystickRead[i]=2; // Mark joystick as trying to read

  // Poll + read joystick, reacquiring if neccesary
  for (j=0;j<2;j++)
  {
    pJoy->Poll();

    nRet=pJoy->GetDeviceState(sizeof(JoyState[i]),&JoyState[i]);
    if (nRet!=DIERR_INPUTLOST) break;
    pJoy->Acquire(); // if DIERR_INPUTLOST, reacquire and try again
  }
  if (nRet<0) return 1;

  // This joystick had been successfully Read this frame:
  bJoystickRead[i]=1;
  return 0;
}

// Read the keyboard
static inline int ReadKeyboard()
{
  int nRet=0,j=0;
  if (Keyboard==NULL) return 1;
  if (bKeyboardRead) return 0; // already read this frame - ready to go

  // Read keyboard, reacquiring if neccesary
  for (j=0;j<2;j++)
  {
    nRet=Keyboard->GetDeviceState(sizeof(KeyboardState),(void *)KeyboardState);
    if (nRet!=DIERR_INPUTLOST) break;
    Keyboard->Acquire(); // if DIERR_INPUTLOST, reacquire and try again
  }
  if (nRet<0) return 1; // Failed to read the keyboard

  // The keyboard has been successfully Read this frame:
  bKeyboardRead=1;
  return 0;
}

// Check a subcode (the 40xx bit in 4001, 4102 etc) for a joystick input code
static int JoystickState(DIJOYSTATE *pState,int nSubCode)
{
  if (nSubCode<5)
  {
    // A joystick direction
    switch (nSubCode)
    {
      case 1: return pState->lY<-0x08000;
      case 2: return pState->lY> 0x08000;
      case 3: return pState->lX<-0x08000;
      case 4: return pState->lX> 0x08000;
    }
    return 0;
  }
  if (nSubCode<0x10) return 0;
  if (nSubCode<0x10 + DINPUT_BUTTON_COUNT) 
  {
    // A joystick button
    return (pState->rgbButtons[nSubCode-0x10]&0x80) ? 1 : 0;
  }
  return 0;
}

// Get the state (pressed=1, not pressed=0) of a particular input code
int DinpState(int nCode)
{
  if (nCode<0) return 0;
  if (nCode<0x100)
  {
    if (ReadKeyboard()!=0) return 0; // Check keyboard has been read - return not pressed on error
    return (KeyboardState[nCode]&0x80)?1:0; // Return key state
  }

  if (nCode<0x4000) return 0;
  if (nCode<0x8000)
  {
    // Codes 4000-8000 = Joysticks
    int nJoyNumber=0;
    nJoyNumber=(nCode-0x4000)>>8;
    if (nJoyNumber>=nJoystickCount) return 0; // This joystick number isn't connected
    if (ReadJoystick(nJoyNumber)!=0) return 0; // There was an error polling the joystick
    // Find the joystick state in our array
    return JoystickState(JoyState+nJoyNumber,nCode&0xff);
  }

  return 0;
}

// This function finds which key is pressed, and returns its code
int DinpFind()
{
  int i=0; int nRetVal=-1; // assume nothing pressed
  DinpStart(); // reset the flags which say each joystick was Read
  // check if any keyboard keys are pressed
  if (ReadKeyboard()==0)
  {
    for (i=0;i<0x100;i++) if (KeyboardState[i]&0x80) { nRetVal=i; goto End; }
  }

  // Now check all the connected joysticks
  for (i=0;i<nJoystickCount;i++)
  {
    int j=0; DIJOYSTATE *pState=NULL;
    if (ReadJoystick(i)!=0) continue; // There was an error polling the joystick
    pState=JoyState+i;
    for (j=1;j<5;j++) { if (JoystickState(pState,j)) { nRetVal=0x4000|(i<<8)|j; goto End; } }
    for (j=0x10;j<0x10+DINPUT_BUTTON_COUNT;j++)
    { if (JoystickState(pState,j)) { nRetVal=0x4000|(i<<8)|j; goto End; } }
  }
End:
  return nRetVal;
}

// Read the axis of a joystick
int DinpAxis(int i,int nAxis)
{
  if (i<0 || i>=nJoystickCount) return 0; // This joystick number isn't connected
  if (ReadJoystick(i)!=0) return 0; // There was an error polling the joystick
  if (nAxis==0) return JoyState[i].lX;
  if (nAxis==1) return JoyState[i].lY;
  if (nAxis==2) return JoyState[i].lZ;
  return 0;
}
